/* $Id: plugins.h,v 2.26 2005/01/16 23:08:08 fknobbe Exp $
 *
 *
 * Copyright (c) 2001-2005 Frank Knobbe <frank@knobbe.us>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 * This file contains the plugins registry.
 *
 */


#ifndef		__PLUGINS_H__
#define		__PLUGINS_H__


#include "snortsam.h"

#ifdef ENABLE_OPSEC
#include "ssp_opsec.h"
#endif

#include "ssp_fwexec.h"
#include "ssp_pix.h"
#include "ssp_ciscoacl.h"
#include "ssp_cisco_nullroute.h"
#include "ssp_email.h"
#include "ssp_netscreen.h"
#include "ssp_ipf.h"
#include "ssp_pf.h"
#include "ssp_ipchains.h"
#include "ssp_iptables.h"
#include "ssp_ebtables.h"
#include "ssp_wgrd.h"
#include "ssp_8signs.h"
#include "ssp_isa.h"
#include "ssp_chxi.h"


#define TH_NONE			0		/* Threading model: See below for more info */
#define TH_SINGLE		1
#define TH_MULTI		2


typedef struct _plugins
{	int (*PluginInit)(DATALIST *);		/* Pointer to init routine, or NULL if not needed */
	void (*PluginConfigParse)(char *,	/* Pointer to config file parsing routine, or NULL if not needed */
		  char *,unsigned long,
		  DATALIST *);
	void (*PluginBlock)(BLOCKINFO *,	/* Pointer to blocking routine, or NULL if not needed (huh?) */
		  void *);				
	void (*PluginExit)(DATALIST *);		/* Pointer to exit routine, or NULL if not needed */
	int PluginNeedsExpiration;			/* Set this to TRUE, if SnortSam needs to expire the block */
	int PluginThreading;				/* Set this to TH_MULTI, TH_SINGLE, TH_NONE. */
	char PluginHandle[40];				/* Short handle, used in config file */
	char PluginAuthor[100];				/* Your name here (or "" if not used) */
	char PluginVersion[30];				/* Version of your plugin (or "" if not used) */
} PLUGINREGISTRY;

/* 
PluginInit:				A function SnortSam calls when it encounters a plugin in the 
						configuration file. The function returns either TRUE or FALSE,
						indicating a successful or unsuccessful initialization. If functions 
						returns FALSE, SnortSam will disable the plugin.
						The paramter is a pointer to the first element in the device/parameter list.
						(Plugin may use it at it's own discretion)

PluginConfigParse:		A function SnortSam calls on every configuration line for the plugin.
						The first parameter is the config line. The second parameter is the
						config file name itself, the third parameter is the line number.
						(the last two are useful for logging purposes.)
						The third parameter is a pointer to a device/parameter structure (not the list).
						If the plugin allocates it, SnortSam will link it into the list for you.

PluginBlock:			A function SnortSam calls when it needs to block an IP address.
						The first parameter is a boolean which is TRUE if the address is to be
						blocked, and FALSE if it is to be unblocked. The second parameter is a
						pointer to the BLOCKINFO struct which contains the IP address, port, 
						protocol, etc.

PluginExit:				A function SnortSam calls when it terminates. This gives the plugin a 
						chance to clean itself up before exiting.
						The paramter is a pointer to the device/parameter list if the plugin uses one.
						(If not, it would still be NULL).
						NOTE: The plugin does not need to free the elements itself anymore. Snortsam will
						free the elements and clean up the chain itself. The parameter is provided for final
						communication to devices if the plugin needs to perform this.

PluginNeedsExpiration:	This can be set to TRUE if SnortSam has to time-out the blocks.
						Can be set to FALSE if the firewall will time-out itself.
						If set to TRUE, SnortSam will keep track of blocks and even create
						a state file so that timeouts/unblocks can be processed even if
						SnortSam is restarted.

PluginThreading:		This can be set to TH_MULTI id the plugin can be executed more than once (for multiple
						firewalls) and can run simultaneously, and is capable if being launched in separate
						threads. The plugin will be launched in parallel with other plugins, and with other
						instances of itself. If the plugin is capable of multithreading, but should only contact
						its own devices sequentially, set this flag to TH_SINGLE. This plugin will then be
						executed in parallel to the other plugins, but it will only process one device at a time.
						If the plugin has problems with threads, set this flag to TH_NONE. In that case SnortSam
						will not launch it in its own thread(s), but run it inline the main thread/process.
						Currently, the OPSEC plugin suffers from this and has to be run that way.
					
PluginHandle:			Short text handle. This is used in config files to pass the config line
						on to the plugin.

PluginAuthor:			The name entered here is listed on startup of SnortSam.

PluginVersion:			The version entered here is listed on startup of SnortSam.

*/

PLUGINREGISTRY Plugins[]={
/* ------------------------------------------------------------ */
/* Native FW-Sam plugin (self assembled packet) */
{	NULL,
	FWSamParse,
	FWSamBlock,
	NULL,
	FALSE,
	TH_MULTI,
	"fwsam",
	"Frank Knobbe",
	"2.2"
},
/* ------------------------------------------------------------ */
/* Old fwexec, now plugin */
{	NULL,
	FWExecParse,
	FWExecBlock,
	NULL,
	FALSE,
	TH_SINGLE,
	"fwexec",
	"Frank Knobbe",
	"2.3"
},
/* ------------------------------------------------------------ */
/* OPSEC compliant plugin */
#ifdef ENABLE_OPSEC
{	NULL,
	OPSEC_Parse,
	OPSEC_Block,
	NULL,
	FALSE,
	TH_NONE,
	"opsec",
	"Frank Knobbe",
	"2.4"
},
#endif
/* ------------------------------------------------------------ */
/* PIX Plugin (using the SHUN command) */
{	NULL,
	PIXParse,
	PIXBlock,
	NULL,
	TRUE,
	TH_MULTI,
	"pix",
	"Frank Knobbe",
	"2.7"
},
/* ------------------------------------------------------------ */
/* CISCO ACL Plugin  */
{	NULL,
	CISCOACLParse,
	CISCOACLBlock,
	NULL,
	TRUE,
	TH_SINGLE,
	"ciscoacl",
	"Ali Basel <alib@sabanciuniv.edu>",
	"2.8"
},
/* ------------------------------------------------------------ */
/* Cisco Null Route Plugin  */
{	NULL,
	CiscoNullRouteParse,
	CiscoNullRouteBlock,
	NULL,
	TRUE,
	TH_MULTI,
	"cisconullroute",
	"Frank Knobbe",
	"2.1"
},
/* ------------------------------------------------------------ */
/* Netscreen Plugin (deny-group) */
{	NULL,
	NetScrnParse,
	NetScrnBlock,
	NULL,
	TRUE,
	TH_MULTI,
	"netscreen",
	"Frank Knobbe",
	"2.7"
},
/* ------------------------------------------------------------ */
#if !defined(WIN32) && !defined(Linux)
/* IPFilter Plugin */
{	NULL,
	IPFParse,
	IPFBlock,
	NULL,
	TRUE,
	TH_SINGLE,
	"ipf",
	"Erik Sneep <erik@webflex.nl>",
	"2.9"
},
#endif
/* ------------------------------------------------------------ */
#ifdef OpenBSD
/* PF Plugin */
{	NULL,
	PFParse,
	PFBlock,
	NULL,
	TRUE,
	TH_MULTI,
	"pf",
	"Hector Paterno <apaterno@dsnsecurity.com>",
 	"3.2"
},
#endif
/* ------------------------------------------------------------ */
#ifdef Linux
/* Ipchains Plugin */
{	NULL,
	IPCHParse,
	IPCHBlock,
	NULL,
	TRUE,
	TH_SINGLE,
	"ipchains",
	"Hector A. Paterno <apaterno@dsnsecurity.com>",
	"2.7"
},   
/* ------------------------------------------------------------ */
/* Iptables Plugin */
{	NULL,
	IPTParse, 
	IPTBlock,   
	NULL,
	TRUE,
	TH_NONE,   /* For now. It may support single-threading */
	"iptables",
	"Fabrizio Tivano <fabrizio@sad.it>",
	"2.6"
},
/* ------------------------------------------------------------ */
/* EBtables Plugin */
{	NULL,
	EBTParse,
	EBTBlock,
	NULL,
	TRUE,
	TH_NONE,   /* For now. It may support single-threading */
	"ebtables",
	"Bruno Scatolin <ipsystems@uol.com.br>",
	"2.2"
},
#endif
/* ------------------------------------------------------------ */
/* Watchguard plugin */
{	NULL,
	WGRDParse,
	WGRDBlock,
	NULL,
	FALSE,
	TH_MULTI,
	"watchguard",
	"Thomas Maier <thomas.maier@arcos.de>",
	"2.3"
},
/* ------------------------------------------------------------ */
#if defined(WIN32)
/* 8signs plugin */
{	NULL,
	DFWParse,
	DFWBlock,
	NULL,
	TRUE,		/* Actually FALSE since it can expire itself. But it only */
	TH_SINGLE,	/* takes day, week, or forever, so we just time-out ourselves. */
	"8signs",
	"Frank Knobbe"
	"2.2"
},
/* ------------------------------------------------------------ */
/* CHX-I plugin */
{	NULL,
	CHXIParse,
	CHXIBlock,
	NULL,
	FALSE,
	TH_SINGLE,
	"chx-i",
	"Frank Knobbe"
	"2.1"
},
/* ------------------------------------------------------------ */
/* Microsoft ISA Server plugin */
{	NULL,
	ISAParse,
	ISABlock,
	NULL,
	TRUE,
	TH_NONE,	/* Maybe single. (COM interfaces calling restrictions) */
	"isa",
	"Nima Sharifi Mehr <nimahacker@yahoo.com>",
	"2.1"
},
#endif
/* ------------------------------------------------------------ */
/* Email Notifcation plugin */
{	NULL,
	EmailParse,
	EmailSend,
	NULL,
	TRUE,
	TH_MULTI,
	"email",
	"Frank Knobbe",
	"2.7"
}
/* ------------------------------------------------------------ */
/* add other plugins here */
};


#endif /* __PLUGINS_H__ */
